package nz.ac.massey.javaecs;
/**
 * Component Manager
 * This class manages component registrations, and adding,
 * removing, and retrieving data from the component arrays.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */
import java.util.Map;
import java.lang.reflect.Type;
import java.util.BitSet;
import java.util.HashMap;

class ComponentManager{
    private Map<Type, ComponentArray> componentArrays = new HashMap<>();
    private Map<Type, Integer> componentPosIndex = new HashMap<>();
    private Map<Integer, Type> indexComponentType = new HashMap<>();

    /**
     * Adds the specified component to the provided entity.
     * <p>
     * Providing a null Object defaults the value to boolean false
     * @param componentType the class type of the component to add
     * @param componentData the component data to associate
     * @param entity the entity to associate data to
     */
    protected boolean addComponentToEntity(Type componentType, Object componentData, Entity entity){
        if (componentData == null){
            // In the case of null component data, insert the boolean false 
            // (preserves structure by associating data with a null component)
            return componentArrays.get(componentType).insertData(entity, false);
        }
        return componentArrays.get(componentType).insertData(entity, componentData);
    }

    /**
     * Signals to the ComponentManager the entity was destroyed. All component data references should be removed.
     * @param entity the entity that was destroyed.
     */
    public void entityDestroyed(Entity entity, BitSet entityRegistrations){
        // HashMap lookups take time, use the known bitstates
        int index = entityRegistrations.nextSetBit(0);
        while(index != -1){
            if (!componentArrays.get(indexComponentType.get(index)).removeData(entity)){
                Engine.writeErr(indexComponentType.get(index).getTypeName());
            }
            index = entityRegistrations.nextSetBit(index+1);
        }
        /*for (Type key : componentArrays.keySet()) {
            if (!componentArrays.get(key).removeData(entity)){
                Engine.writeErr(key.getTypeName());
            }
        }*/
    }

    /**
     * Gets the component data associated with the entity
     * @param componentType the class type of data to look for
     * @param entity the entity to find data for
     * @return the Object data found, or null if it was not found
     */
    public Object getComponent(Type componentType, Entity entity){
        return componentArrays.get(componentType).getData(entity);
    }

    /**
     * Gets the registration index of the component type
     * @param type the class type of the component
     * @return the index of the component type, or -1 if it isn't found
     */
    protected Integer getComponentIndex(Type type){
        try{
            return componentPosIndex.get(type);
        }
        catch (NullPointerException e){
            return -1;
        }        
    }

    /**
     * Gets the type of the component at the provided index
     * @param index the index of the component
     * @return the class type of the index. `null` if not found
     */
    protected Type getComponentType(Integer index){
        try{
            return indexComponentType.get(index);
        }
        catch (NullPointerException e){
            return null;
        }  
    }

    /**
     * Moves a single component data from one entity to another
     * @param sourceEntity the entity to move data from
     * @param destinationEntity the entity to move data to
     * @param component the component class type to consider 
     * @return true if the component was moved successfully, else false
     */
    protected boolean moveComponentData(Entity sourceEntity, Entity destinationEntity, Type component){
        if (componentArrays.get(component).moveData(sourceEntity, destinationEntity) == 0){
            return true;
        }
        else return false;
    }

    /**
     * Moves all component data from one entity to another
     * @param sourceEntity the entity to move data from
     * @param destinationEntity the entity to move data to
     * @param sourceRegistrations the component registrations of the source entity
     */
    protected void moveAllComponentData(Entity sourceEntity, Entity destinationEntity, BitSet sourceRegistrations){
        int result = sourceRegistrations.nextSetBit(0);
        while (result != -1){
            Type key = indexComponentType.get(result);
            componentArrays.get(key).moveData(sourceEntity, destinationEntity);
            result = sourceRegistrations.nextSetBit(result+1);
        }
    }

    /**
     * Registers the component type
     * @param type the class type to register
     * @return true if the component was registered successfully, else false
     */
    protected boolean registerComponent(Type type){
        if (componentArrays.containsKey(type)){
            Engine.writeErr("Component " + type.getTypeName() + " is already registered");
            return false;
        }
        componentArrays.put(type, new ComponentArray());
        indexComponentType.put(componentPosIndex.size(), type);
        componentPosIndex.put(type, componentPosIndex.size());
        return true;
    }

    /**
     * Removes the specified component from the entity
     * @param componentType the class type of the component to remove
     * @param entity the entity to remove the component from
     */
    public boolean removeComponentFromEntity(Type componentType, Entity entity){
        return componentArrays.get(componentType).removeData(entity);
    }
}