package nz.ac.massey.javaecs;

import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.BitSet;

import org.junit.jupiter.api.BeforeEach;

/**
 * Unit test for simple App.
 */
class AppTest {

    Engine gameEngine;
    TestSystem system;

    class TestObject{
        TestObject() {}
        TestObject(int i){
            this.i = i;
        }
        int i = 1;
    }

    class TestSystem extends ECSSystem{
        public void init(){}
        public void update() {
            for (Entity entity : entities) {
                TestObject entityComponent = (TestObject)gameEngine.getComponentData(entity, TestObject.class);
                entityComponent.i++;
            }
        }
    }
    /**
     * Test assigning an entity
     */
    @Test
    void testAssignOne() {
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());
        assertEquals(1, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
    }

    /**
     * Test assigning the maximum number of entities
     */
    @Test
    void testAssignMax(){
        for (int i = 0; i < gameEngine.getMaxEntities(); i++) {
            Entity entity = gameEngine.createEntity();
            gameEngine.addComponent(entity, TestObject.class, new TestObject());
            assertEquals(1, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
        }
    }
    /**
     * Test assigning more than the maximum number of entities
     */
    @Test
    void testAssignMoreThanMax(){
        for (int i = 0; i < gameEngine.getMaxEntities() + 5; i++) {
            if (i >= gameEngine.getMaxEntities()){
                assertThrows(IndexOutOfBoundsException.class, () -> {
                    gameEngine.createEntity();
                });
            }
            else{
                Entity entity = gameEngine.createEntity();
                gameEngine.addComponent(entity, TestObject.class, new TestObject());
            }
        }
    }

    /**
     * Test removing data
     */
    @Test
    void testRemoveOne(){
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());
        gameEngine.destroyEntity(entity);
        assertNull(((TestObject)gameEngine.getComponentData(entity, TestObject.class)));
    }

    /**
     * Test running the system several times
     */
    @Test
    void testRunSystem(){
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());

        for (int i = 0; i < 5; i++) {
            system.update();
        }

        assertEquals(6, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
    }


    /**
     * Test resizing (minimal, with only 1 entity)
     */
    @Test
    void testResizeMinimal(){
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());

        gameEngine.resizeMaximum(10);

        assertEquals(10, gameEngine.getMaxEntities());
    }

    /**
     * Test large creations & deletions
     */
    @Test
    void testLargeCreateDelete(){
        for (int i = 0; i < gameEngine.getMaxEntities(); i++) {
            Entity entity = gameEngine.createEntity();
            gameEngine.addComponent(entity, TestObject.class, new TestObject());
            assertEquals(1, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
        }
        for (int i = 256; i < 512; i++) {
            gameEngine.destroyEntity(Entity.asEntity(i));
        }
        int k = 0;
    }

    /**
     * Test large creations & deletions & smaller resize
     */
    @Test
    void testLargeCreateDeleteResize(){
        for (int i = 0; i < gameEngine.getMaxEntities(); i++) {
            Entity entity = gameEngine.createEntity();
            gameEngine.addComponent(entity, TestObject.class, new TestObject());
            assertEquals(1, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
        }
        for (int i = 256; i < 512; i++) {
            gameEngine.destroyEntity(Entity.asEntity(i));
        }

        gameEngine.resizeMaximum(1024 - 255);
    }

    /**
     * Test large creations & deletions & smaller resize, to a value too small
     */
    @Test
    void testLargeCreateDeleteResizeTooSmall(){
        for (int i = 0; i < gameEngine.getMaxEntities(); i++) {
            Entity entity = gameEngine.createEntity();
            gameEngine.addComponent(entity, TestObject.class, new TestObject());
            assertEquals(1, ((TestObject)gameEngine.getComponentData(entity, TestObject.class)).i);
        }
        for (int i = 256; i < 512; i++) {
            gameEngine.destroyEntity(Entity.asEntity(i));
        }
        assertFalse(gameEngine.resizeMaximum(512));
        
    }

    @Test
    void testEntityManagerConstructor(){
        gameEngine = new Engine(5);
        assertEquals(5, gameEngine.getMaxEntities());
    }

    @Test
    void testAssignToNonExistentEntity(){
        assertFalse(gameEngine.addComponent(Entity.asEntity(1025), TestObject.class, new TestObject()));
    }

    @Test
    void testRemoveComponent(){
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());
        gameEngine.removeComponent(entity, TestObject.class);
        assertFalse(gameEngine.entityManager.getRegistrations(entity).get(gameEngine.getComponentIndex(TestObject.class)));
    }

    @Test
    void testEqualResize(){
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        PrintStream orig = Engine.getErr();
        PrintStream newErr = new PrintStream(bytes);
        Engine.setErr(newErr);
        assertTrue(gameEngine.resizeMaximum(1024));
        Engine.setErr(orig);
        newErr.flush(); // ensure the bytes are recieved
        byte[] errBytes = bytes.toByteArray();
        String result = new String(errBytes);
        Engine.writeErr("Captured in redirect: " + result);
        assertTrue(result.trim().equals("Attempted to set the newSize to the current size"));
    }

    @Test
    void testLargerResize(){
        assertTrue(gameEngine.resizeMaximum(2048));
        assertEquals(2048, gameEngine.getMaxEntities());
    }

    @Test
    void testComponentAlreadyRegistered(){
        assertFalse(gameEngine.registerComponent(TestObject.class));
    }

    @Test
    void testRemoveNonExistentComponentData(){
        assertFalse(gameEngine.removeComponent(Entity.asEntity(1), TestObject.class));
    }

    @Test
    void testAssignComponentAlreadyAssigned(){
        Entity entity = gameEngine.createEntity();
        gameEngine.addComponent(entity, TestObject.class, new TestObject());
        assertFalse(gameEngine.addComponent(entity, TestObject.class, new TestObject()));
    }

    @Test
    void testRegisterExistingSystem(){
        assertFalse(gameEngine.registerSystem(TestSystem.class, new TestSystem()));
    }

    @Test
    void testGetRegistrationsOutOfRange(){
        Entity entity = Entity.asEntity(-1);
        assertEquals(new BitSet(), gameEngine.entityManager.getRegistrations(entity));
    }

    @Test
    void testGetRegistrationsUnassignedEntity(){
        Entity entity = Entity.asEntity(25);
        assertEquals(new BitSet(), gameEngine.entityManager.getRegistrations(entity));
    }

    /**
     * Establish a simple ECS, with a single system and component
     */
    @BeforeEach
    void testSetup(){
        gameEngine = new Engine();
        gameEngine.registerComponent(TestObject.class);

        system = new TestSystem();
        gameEngine.registerSystem(TestSystem.class, system);
        {
            BitSet signature = new BitSet();
            signature.set(gameEngine.getComponentIndex(TestObject.class));
            gameEngine.setSystemSignature(TestSystem.class, signature);
        }

        system.init();
    }
}
